import equadratures.sampling_methods.sampling_template
import equadratures.sampling_methods.montecarlo
import equadratures.sampling_methods.tensorgrid
import equadratures.sampling_methods.sparsegrid
import equadratures.sampling_methods.userdefined
import equadratures.sampling_methods.induced